/**
 * Copyright (C) 2011-2012 enStratus Networks Inc
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.jclouds.cloudsigma.compute;

import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Locale;

import org.apache.log4j.Logger;
import org.dasein.cloud.AsynchronousTask;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.CloudProvider;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.OperationNotSupportedException;
import org.dasein.cloud.compute.Architecture;
import org.dasein.cloud.compute.MachineImage;
import org.dasein.cloud.compute.MachineImageFormat;
import org.dasein.cloud.compute.MachineImageState;
import org.dasein.cloud.compute.MachineImageSupport;
import org.dasein.cloud.compute.MachineImageType;
import org.dasein.cloud.compute.Platform;
import org.dasein.cloud.identity.ServiceAction;
import org.dasein.cloud.jclouds.cloudsigma.CloudSigma;
import org.jclouds.cloudsigma.CloudSigmaAsyncClient;
import org.jclouds.cloudsigma.CloudSigmaClient;
import org.jclouds.cloudsigma.domain.DriveInfo;
import org.jclouds.cloudsigma.domain.DriveType;
import org.jclouds.rest.RestContext;

import javax.annotation.Nonnull;

public class CSImage implements MachineImageSupport {
    static private final Logger logger = Logger.getLogger(CSImage.class);

    private CloudSigma cloud;
    
    CSImage(CloudSigma cloud) { this.cloud = cloud; }
    
    @Override
    public void downloadImage(String machineImageId, OutputStream toOutput) throws CloudException, InternalException {
        throw new OperationNotSupportedException("Downloads not currently supported");
    }

    DriveInfo getDrive(String machineImageId) throws CloudException, InternalException {
        RestContext<CloudSigmaClient,CloudSigmaAsyncClient> ctx = cloud.getCloudClient();
        
        try {
            try {
                return ctx.getApi().getDriveInfo(machineImageId);
            }
            catch( Exception e ) {
                logger.error("Could not get drive info for " + machineImageId + ": " + e.getMessage());
                if( logger.isDebugEnabled() ) {
                    e.printStackTrace();
                }
                throw new CloudException(e);
            }
        }
        finally {
            ctx.close();
        }        
    }
    
    @Override
    public MachineImage getMachineImage(String machineImageId) throws CloudException, InternalException {
        return toMachineImage(getDrive(machineImageId));
    }

    @Override
    public String getProviderTermForImage(Locale locale) {
        return "boot volume";
    }

    @Override
    public boolean hasPublicLibrary() {
        return true;
    }

    @Override
    public AsynchronousTask<String> imageVirtualMachine(String vmId, String name, String description) throws CloudException, InternalException {
        // TODO Auto-generated method stub
        return null;
    }

    @Override
    public AsynchronousTask<String> imageVirtualMachineToStorage(String vmId, String name, String description, String directory) throws CloudException, InternalException {
        // TODO Auto-generated method stub
        return null;
    }

    @Override
    public String installImageFromUpload(MachineImageFormat format, InputStream imageStream) throws CloudException, InternalException {
        // TODO Auto-generated method stub
        return null;
    }

    @Override
    public boolean isImageSharedWithPublic(String machineImageId) throws CloudException, InternalException {
        return false; // TODO: really?
    }

    @Override
    public boolean isSubscribed() throws CloudException, InternalException {
        RestContext<CloudSigmaClient,CloudSigmaAsyncClient> ctx = cloud.getCloudClient();
        
        try {
            try {
                ctx.getApi().listDrives();
                return true;
            }
            catch( Exception e ) {
                if( e.getMessage().contains("errors:user") || e.getMessage().contains("errors:password") ) {
                    return false;
                }
                logger.warn("Could not check subscription status in CloudSigma for " + cloud.getContext().getAccountNumber() + ": " + e.getMessage());
                if( logger.isDebugEnabled() ) {
                    e.printStackTrace();
                }
                throw new CloudException(e);
            }
        }
        finally {
            ctx.close();
        }  
    }

    @Override
    public Iterable<MachineImage> listMachineImages() throws CloudException, InternalException {
        return listMachineImagesOwnedBy(cloud.getContext().getAccountNumber());
    }

    @Override
    public Iterable<MachineImage> listMachineImagesOwnedBy(String accountId) throws CloudException, InternalException {
        try {
            if( accountId == null ) {
                RestContext<CloudSigmaClient,CloudSigmaAsyncClient> ctx = cloud.getCloudClient();
                
                try {
                    ArrayList<MachineImage> images = new ArrayList<MachineImage>();
                    Collection<String> ids;
                    
                    try {
                        ids = ctx.getApi().listStandardDrives();
                    }
                    catch( RuntimeException e ) {
                        logger.warn("Error listing standard drives: " + e.getMessage());
                        if( logger.isDebugEnabled() ) {
                            e.printStackTrace();
                        }
                        throw new CloudException(e);
                    }
                    for( String driveId : ids ) {
                        DriveInfo drive;
                        MachineImage img;
                        
                        try {
                            drive = ctx.getApi().getDriveInfo(driveId);
                        }
                        catch( RuntimeException e ) {
                            logger.warn("Error getting drive info for " + driveId + ": " + e.getMessage());
                            if( logger.isDebugEnabled() ) {
                                e.printStackTrace();
                            }
                            throw new CloudException(e);                        
                        }
                        img = toMachineImage(drive);
                        if( img != null ) {
                            images.add(img);
                        }
                    }
                    return images;
                }
                finally {
                    ctx.close();
                }
            }
            else if( accountId.equals(cloud.getContext().getAccountNumber()) ) {
                RestContext<CloudSigmaClient,CloudSigmaAsyncClient> ctx = cloud.getCloudClient();
                
                try {
                    ArrayList<MachineImage> images = new ArrayList<MachineImage>();
                    Collection<String> ids;
                    
                    try {
                        ids = ctx.getApi().listDrives();
                    }
                    catch( RuntimeException e ) {
                        logger.warn("Error listing private drives: " + e.getMessage());
                        if( logger.isDebugEnabled() ) {
                            e.printStackTrace();
                        }
                        throw new CloudException(e);                    
                    }
                    for( String driveId : ids ) {
                        DriveInfo drive;
                        MachineImage img;
                        
                        try {
                            drive = ctx.getApi().getDriveInfo(driveId);
                        }
                        catch( RuntimeException e ) {
                            logger.warn("Error getting drive info for " + driveId + ": " + e.getMessage());
                            if( logger.isDebugEnabled() ) {
                                e.printStackTrace();
                            }
                            throw new CloudException(e);                            
                        }
                        img = toMachineImage(drive);
                        if( img != null ) {
                            images.add(img);
                        }
                    }
                    return images;
                }
                finally {
                    ctx.close();
                }            
            }
            else {
                return Collections.emptyList();
            }
        }
        catch( RuntimeException e ) {
            logger.error("Runtime error during image listing: " + e.getMessage());
            if( logger.isDebugEnabled() ) {
                e.printStackTrace();
            }
            throw new InternalException(e);
        }
        catch( Error e ) {
            logger.error("Error during image listing: " + e.getMessage());
            if( logger.isDebugEnabled() ) {
                e.printStackTrace();
            }
            throw new InternalException(e);
        }        
    }

    @Override
    public Iterable<MachineImageFormat> listSupportedFormats() throws CloudException, InternalException {
        return Collections.emptyList();
    }

    @Override
    public Iterable<String> listShares(String forMachineImageId) throws CloudException, InternalException {
        return Collections.emptyList();
    }

    @Override
    public @Nonnull String[] mapServiceAction(@Nonnull ServiceAction action) {
        return new String[0];
    }

    @Override
    public String registerMachineImage(String atStorageLocation) throws CloudException, InternalException {
        // TODO Auto-generated method stub
        return null;
    }

    @Override
    public void remove(String machineImageId) throws CloudException, InternalException {
        // TODO Auto-generated method stub
        
    }

    @Override
    public Iterable<MachineImage> searchMachineImages(String keyword, Platform platform, Architecture architecture) throws CloudException, InternalException {
        try {
            ArrayList<MachineImage> list = new ArrayList<MachineImage>();
            Iterable<MachineImage> raw;
            
            try {
                raw = listMachineImages();
            }
            catch( RuntimeException e ) {
                logger.warn("Error listing private images for search: " + e.getMessage());
                if( logger.isDebugEnabled() ) {
                    e.printStackTrace();
                }
                throw new CloudException(e);
            }
            for( MachineImage image :  raw ) {
                if( architecture != null && !architecture.equals(image.getArchitecture()) ) {
                    continue;
                }
                if( platform != null && !platform.equals(Platform.UNKNOWN) ) {
                    Platform mine = image.getPlatform();
                    
                    if( platform.isWindows() && !mine.isWindows() ) {
                        continue;
                    }
                    if( platform.isUnix() && !mine.isUnix() ) {
                        continue;
                    }
                    if( platform.isBsd() && !mine.isBsd() ) {
                        continue;
                    }
                    if( platform.isLinux() && !mine.isLinux() ) {
                        continue;
                    }
                    if( platform.equals(Platform.UNIX) ) {
                        if( !mine.isUnix() ) {
                            continue;
                        }
                    }
                    else if( !platform.equals(mine) ) {
                        continue;
                    }
                }
                if( keyword != null && !keyword.equals("") ) {
                    keyword = keyword.toLowerCase();
                    if( !image.getProviderMachineImageId().toLowerCase().contains(keyword) ) {
                        if( !image.getName().toLowerCase().contains(keyword) ) {
                            if( !image.getDescription().toLowerCase().contains(keyword) ) {
                                continue;
                            }
                        }
                    }
                }
                list.add(image);
            }
            try {
                raw = listMachineImagesOwnedBy(null);
            }
            catch( RuntimeException e ) {
                logger.warn("Error listing public images for search: " + e.getMessage());
                if( logger.isDebugEnabled() ) {
                    e.printStackTrace();
                }
                throw new CloudException(e);            
            }
            for( MachineImage image : raw ) {
                if( architecture != null && !architecture.equals(image.getArchitecture()) ) {
                    continue;
                }
                if( cloud.getContext().getAccountNumber().equals(image.getProviderOwnerId()) ) {
                    continue;
                }
                if( platform != null && !platform.equals(Platform.UNKNOWN) ) {
                    Platform mine = image.getPlatform();
                    
                    if( platform.isWindows() && !mine.isWindows() ) {
                        continue;
                    }
                    if( platform.isUnix() && !mine.isUnix() ) {
                        continue;
                    }
                    if( platform.isBsd() && !mine.isBsd() ) {
                        continue;
                    }
                    if( platform.isLinux() && !mine.isLinux() ) {
                        continue;
                    }
                    if( platform.equals(Platform.UNIX) ) {
                        if( !mine.isUnix() ) {
                            continue;
                        }
                    }
                    else if( !platform.equals(mine) ) {
                        continue;
                    }
                }
                if( keyword != null && !keyword.equals("") ) {
                    keyword = keyword.toLowerCase();
                    if( !image.getProviderMachineImageId().toLowerCase().contains(keyword) ) {
                        if( !image.getName().toLowerCase().contains(keyword) ) {
                            if( !image.getDescription().toLowerCase().contains(keyword) ) {
                                continue;
                            }
                        }
                    }
                }
                list.add(image);
            }
            return list;
        }
        catch( RuntimeException e ) {
            logger.error("Runtime error during image search: " + e.getMessage());
            if( logger.isDebugEnabled() ) {
                e.printStackTrace();
            }
            throw new InternalException(e);
        }
        catch( Error e ) {
            logger.error("Error during image search: " + e.getMessage());
            if( logger.isDebugEnabled() ) {
                e.printStackTrace();
            }
            throw new InternalException(e);
        }        
    }

    @Override
    public void shareMachineImage(String machineImageId, String withAccountId, boolean allow) throws CloudException, InternalException {
        throw new OperationNotSupportedException("No machine image sharing in CloudSigma.");
    }

    @Override
    public boolean supportsCustomImages() {
        return false;
    }

    @Override
    public boolean supportsImageSharing() {
        return false;
    }

    @Override
    public boolean supportsImageSharingWithPublic() {
        return false;
    }

    @Override
    public String transfer(CloudProvider fromCloud, String machineImageId) throws CloudException, InternalException {
        // TODO Auto-generated method stub
        return null;
    }

    private MachineImage toMachineImage(DriveInfo drive) {
        if( drive != null && drive.getType().equals(DriveType.DISK) && drive.getDriveType().contains("preinstalled") ) {
            MachineImage image = new MachineImage();
            
            image.setProviderOwnerId(drive.getUser());
            image.setProviderRegionId(cloud.getContext().getRegionId());
            image.setProviderMachineImageId(drive.getUuid());
            image.setName(drive.getName());
            if( image.getName() == null ) {
                image.setName(image.getProviderMachineImageId());
            }
            image.setDescription(image.getDescription());
            if( image.getDescription() == null ) {
                image.setDescription(image.getName());
            }
            if( drive.getBits() == 32 ) {
                image.setArchitecture(Architecture.I32);
            }
            else {
                image.setArchitecture(Architecture.I64);
            }
            image.setPlatform(Platform.guess(drive.getOs()));
            if( image.getPlatform().equals(Platform.UNKNOWN) ) {
                image.setPlatform(Platform.guess(drive.getName()));
            }
            switch( drive.getStatus() ) {
            case ACTIVE: image.setCurrentState(MachineImageState.ACTIVE); break;
            case INACTIVE: case UNRECOGNIZED: image.setCurrentState(MachineImageState.DELETED); break;
            default: image.setCurrentState(MachineImageState.PENDING); break;
            }
            image.setType(MachineImageType.VOLUME);
            image.setSoftware("");
            return image;
        }
        return null;
    }
}
